#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

#include "autothrottle.h"
#include "config.h"
#include "msg.h"
#include "acpi.h"
#include "cpu.h"

static int current_state;
static scale *states;
static int state_count;

int main(int argc, char *argv[]) {
#ifdef DEBUG
	int i = 0;
#endif
	if(get_config(argc, argv) == ERRVAL)
		return EXIT_FAILURE;

	if(build_state_array() == ERRVAL)
		return EXIT_FAILURE;
#ifdef DEBUG
	printf("Dumping array of throttling states...\n");
	for(i = 0; i < state_count; i++)
		printf("State %i: up %i down %i\n", i, states[i].up, states[i].down);
#endif

#ifdef TEST
	if(do_test) {	
		config.daemonize = 0;
		if(test() == ERRVAL) {
			printf("Test oopsed. Check your settings.\n");
			return EXIT_FAILURE;
		} else {
			printf("Test complete. No errors. Everything went just fine. Great.\n");
			return EXIT_SUCCESS;
		}
	}
#endif

	current_state = state_count - 1;

	if(enable_userspace_governor() == ERRVAL) 
		return EXIT_FAILURE;

	if(config.daemonize == 0) {
		if(config.loglevel == LOG_NOTE) msg("Running in foreground.", LOG_NOTE);
	} else {
		if(config.loglevel == LOG_NOTE) msg("Daemonizing...", LOG_NOTE);
		if(daemon(0, 0) == -1) {
			msg("Unable to fork to daemon.", LOG_ERROR);
			return EXIT_FAILURE;
		}
	}

	signal(SIGHUP, catch_int);

	while(1) {
		usleep(100000);
		if(check() == ERRVAL)
			return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}

void catch_int(int sig_num) {
	signal(SIGHUP, catch_int);
	check_ac();
	if(build_state_array() == ERRVAL)
		exit(EXIT_FAILURE);
}

int test(void) {
	int i = 0, j = 0, step = (config.max_cpu_speed - config.min_cpu_speed) / (config.cpu_states - 1);
	int acpi = 0, cpu = 0;
	config.loglevel = LOG_NOTE;

	printf("OK, here we go, let's perform some tests to see if we can set CPU and ACPI states.\n");

	printf("Enabling userspace governor...\n");
	if(enable_userspace_governor() == ERRVAL) return ERRVAL;

	printf("Saving current throttling settings...\n");
	if((acpi = get_acpi()) == ERRVAL) return ERRVAL;
	if((cpu = get_speed()) == ERRVAL) return ERRVAL;

	/* Test CPU speeds */
	for(i = config.min_cpu_speed; i <= config.max_cpu_speed; i += step) {
		if(set_speed(i) == ERRVAL) return ERRVAL;
		if(i != get_speed()) {
			printf("Ooops. Speed set, but didn't change.\n");
			return ERRVAL;
		} else {
			printf("Speed set to %i ok.\n", i);
		}
		/* Test all ACPI states for every CPU setting */
		for(j = config.acpi_states - 1; j >= 0; j--) {
			if(set_acpi(j) == ERRVAL) return ERRVAL;
			if(j != get_acpi()) {
				printf("Ooops. State set, but didn't change.\n");
				return ERRVAL;
			} else {
				printf("ACPI set to %i ok.\n", j);
			}
		}
	}

	printf("Restoring settings...\n");
	if(set_acpi(acpi) == ERRVAL) {
		return ERRVAL;
	}
	if(set_speed(cpu) == ERRVAL) {
		return ERRVAL;
	}

	return OKVAL;
}
	
int build_state_array(void) {
	int i = 0, j = 0, cpu_speed_step = (config.max_cpu_speed - config.min_cpu_speed) / (config.cpu_states - 1);

	states = (scale *) calloc(config.cpu_states + config.acpi_states - 1, sizeof(scale));
	if(states == NULL) {
		msg("Could not allocate memory.", LOG_ERROR);
		return ERRVAL;
	}

	state_count = 0;

	/* Fill with acpi throttling states */
	if(config.use_acpi) {
		for(i = config.acpi_states - 1; i > 0; i--) {
			/* Within limits? */
			if(((1 - (i * (1 / (double) config.acpi_states))) * config.min_cpu_speed) < config.min_speed) continue;

			states[j].up = i - 1;
			if(i < (config.acpi_states - 1)) states[j].down = i + 1;
			else states[j].down = -1;
			j++;
			state_count++;
		}
	}

	/* Fill with cpufreq throttling states */
	if(config.use_cpufreq) {
		for(i = config.min_cpu_speed; i <= config.max_cpu_speed; i += cpu_speed_step) {
			/* Are we within the limits? */
			if(i > config.max_speed) break;
			else if(i < config.min_speed) continue;

			if(i == config.max_cpu_speed) states[j].up = -1;
			else states[j].up = i + cpu_speed_step;
			if(i > config.min_cpu_speed) states[j].down = i - cpu_speed_step;
			else states[j].down = 1;
			j++;
			state_count++;
		}
	}

	states = (scale *) realloc((void *) states, state_count * sizeof(scale));
	if(states == NULL) {
		msg("Could not allocate memory.", LOG_ERROR);
		return ERRVAL;
	}

	return OKVAL;
}

int check(void) {
	double load = get_cpu_load();
	if(load == -1.0) return ERRVAL;

	if((load >= config.up_threshold) && (current_state != (state_count - 1))) {
		if(up() == ERRVAL) return ERRVAL;
	}
	else if((load < config.down_threshold) && (current_state != 0)) {
		if(down() == ERRVAL) return ERRVAL;
	}

	return OKVAL;
}

int up(void) {
	if(states[current_state].up > config.acpi_states) {
		if(set_speed(states[current_state].up) == ERRVAL) return ERRVAL;
	} else {
		if(set_acpi(states[current_state].up) == ERRVAL) return ERRVAL;
	}
	current_state++;
	return OKVAL;
}

int down(void) {
	if(states[current_state].down > config.acpi_states) {
		if(set_speed(states[current_state].down) == ERRVAL) return ERRVAL;
	} else {
		if(set_acpi(states[current_state].down) == ERRVAL) return ERRVAL;
	}
	current_state--;
	return OKVAL;
}
